from ami2rabbitmq.entities import (
    Bridge,
    Endpoint,
    QueueCaller,
    QueueMember,
    Queue,
    get_endpoint_key,
    get_queue_key
)
from ami2rabbitmq.redis_db import redis_server
from ami2rabbitmq.exceptions import InvalidEvent


class BaseBridgeRepositorie:

    def __init__(self) -> None:
        self._changed_objects = []
        self._redis_server = redis_server
    
    def clear(self) -> None:
        self._changed_objects = []
    
    def get_entity(self, event: dict):
        raise NotImplementedError()
    
    def create(self, event: dict):
        entity = self.entity_create(event)
        self._changed_objects.append(entity)
        self._redis_server.save(
            entity.get_key(), 
            entity.get_dict()
        )
        return entity
    
    def delete(self, event: dict):
        entity = self.entity_delete(event)
        self._changed_objects.append(entity)
        self._redis_server.delete(
            entity.get_key()
        )
        return entity
        

class BaseEndpointQueueRepositorie:

    def __init__(self) -> None:
        self._redis_server = redis_server
        self._changed_objects = []

    def get_entity(self, event: dict):
        raise NotImplementedError()
    
    def clear(self):
        self._changed_objects = []

    def create(self, event: dict):
        entity = self.get_entity(event)
        self._changed_objects.append(entity)
        redis_server.save(
            entity.get_key(), 
            entity.get_dict()
        )
        return entity


class Bridges(BaseBridgeRepositorie):

    def _get_endpoint_key(self, bridge):
        return get_endpoint_key(bridge.device)
    
    def _get_endpoint_data(self, bridge):
        endpoint_key = self._get_endpoint_key(bridge)

        if self._redis_server.exists(endpoint_key):
            return self._redis_server.get(endpoint_key)
        
        return None

    def _get_entity(self, event: dict) -> Bridge:
        if event['Event'] in ['BridgeEnter', 'BridgeLeave'] :
            return  Bridge(event)
        
        else:
            raise InvalidEvent(f"Event Invalid: {event['Event']}")
    
    def entity_create(self, event: dict) -> Bridge:
        bridge = self._get_entity(event)
        endpoint_data = self._get_endpoint_data(bridge)

        if endpoint_data:
            endpoint_data['bridges'].append(bridge.uniqueid)
            self._redis_server.save(
                self._get_endpoint_key(bridge), 
                endpoint_data
            )
        
        return bridge
    
    def entity_delete(self, event: dict) -> Bridge:
        bridge = self._get_entity(event)
        endpoint_data = self._get_endpoint_data(bridge)

        if endpoint_data:
            endpoint_data['bridges'].remove(bridge.uniqueid)
            self._redis_server.save(
                self._get_endpoint_key(bridge), 
                endpoint_data
            )
        
        return bridge


class Endpoints(BaseEndpointQueueRepositorie):

    def _get_endpoint_bridges(self, endpoint: Endpoint) -> list:
        if self._redis_server.exists(endpoint.get_key()):
            return self._redis_server.get(endpoint.get_key())['bridges']
        
        return []
 
    def _endpoint_create(self, event: dict) -> Endpoint:
        endpoint = Endpoint(event)
        endpoint.bridges = self._get_endpoint_bridges(endpoint)       
        return endpoint
    
    def get_entity(self, event: dict) -> Endpoint:

        if event['Event'] == 'DeviceStateChange':
            if 'Queue:' not in event['Device']:
                return self._endpoint_create(event)
        
        raise InvalidEvent(f"Event Invalid: {event['Event']}")


class QueueCallers(BaseBridgeRepositorie):

    def _get_queue_key(self, queue_caller: QueueCaller):
       return get_queue_key(queue_caller.device)
    
    def _get_queue_data(self, queue_caller: QueueCaller):
        queue_key = self._get_queue_key(queue_caller)

        if self._redis_server.exists(queue_key):
            return self._redis_server.get(queue_key)
        
        return None

    def _get_entity(self, event: dict) -> QueueCaller:
        if event['Event'] in ['QueueCallerJoin', 'QueueCallerLeave']:
            return QueueCaller(event)
        
        else:
            raise InvalidEvent(f"Event Invalid: {event['Event']}")
    
    def entity_create(self, event: dict) -> QueueCaller:
        queue_caller = self._get_entity(event)
        queue_data = self._get_queue_data(queue_caller)

        if queue_data:
            queue_data['queue_callers'].append(queue_caller.uniqueid)
            self._redis_server.save(
                self._get_queue_key(queue_caller), 
                queue_data
            )
        
        return queue_caller
    
    def entity_delete(self, event: dict) -> QueueCaller:
        queue_caller = self._get_entity(event)
        queue_data = self._get_queue_data(queue_caller)

        if queue_data:
            queue_data['queue_callers'].remove(queue_caller.uniqueid)
            self._redis_server.save(
                self._get_queue_key(queue_caller), 
                queue_data
            )
        
        return queue_caller


class QueueMembers(BaseEndpointQueueRepositorie):

    def _get_queues_list(self, queue_member: QueueMember)-> list:
        if self._redis_server.exists(queue_member.get_list_queues_key()):
            return self._redis_server.get(queue_member.get_list_queues_key())

        return []

    def _queue_member_create(self, event: dict) -> QueueMember:
        queue_member = QueueMember(event)
        queue_member.queues_list = self._get_queues_list(queue_member)

        if event['Queue'] not in queue_member.queues_list:
            queue_member.queues_list.append(event['Queue'])
            self._redis_server.save(queue_member.get_list_queues_key(), queue_member.queues_list)

        return queue_member

    def get_entity(self, event: dict) -> QueueMember:
        if event['Event'] == 'QueueMemberStatus':
            return self._queue_member_create(event)
        
        raise InvalidEvent(f"Event Invalid: {event['Event']}")


class Queues(BaseEndpointQueueRepositorie):

    def _get_queue_callers(self, queue: Queue) -> list:
        if self._redis_server.exists(queue.get_key()):
            return self._redis_server.get(queue.get_key())['queue_callers'] 

        return [] 
    
    def _queue_create(self, event: dict) -> Queue:
        queue = Queue(event)
        queue.queue_callers = self._get_queue_callers(queue)
        return queue

    def get_entity(self, event: dict) -> Queue:
        if event['Event'] == 'DeviceStateChange':
            if 'Queue:' in event['Device']:
                return self._queue_create(event)
        
        raise InvalidEvent(f"Event Invalid: {event['Event']}")


from ami2rabbitmq.ami_rabbitmq import AMI2RabbitMQ
import ami2rabbitmq.entities as entities
import ami2rabbitmq.exceptions as exceptions
import ami2rabbitmq.logs_app as logs_app
import ami2rabbitmq.manager as manager
import ami2rabbitmq.rabbitmq as rabbitmq
import ami2rabbitmq.redis_db as redis_db
import ami2rabbitmq.repositories as repositories
import ami2rabbitmq.utils as utils


__all__ = [
    AMI2RabbitMQ,
    entities,
    exceptions,
    logs_app,
    manager,
    rabbitmq,
    redis_db,
    repositories,
    utils
]


import sys
from time import sleep
from asterisk.manager import (
    Manager,
    ManagerSocketException,
    ManagerAuthException,
    ManagerException
)
from ami2rabbitmq.manager import Pabx
from ami2rabbitmq.rabbitmq import RabbitMQProducer
from ami2rabbitmq.logs_app import LogsApp


class AMI2RabbitMQ():
    _manager = Manager()
    _pabx = Pabx()
    status = None
    last_events = []

    def __init__(self, ami_settings: dict, rabbitmq_settings: dict, debug: bool=False) -> None:
        self._logs_app = LogsApp(debug)
        self._rabbitmq_producer = RabbitMQProducer(rabbitmq_settings, self._logs_app)
        self.events = ami_settings['events']
        self.host = ami_settings['host']
        self.user = ami_settings['user']
        self.password = ami_settings['password']

    def _connect(self) -> None:
        self._manager.connect(self.host)
        self._manager.login(self.user, self.password)
    
    def _disconnect(self) -> None:
        self._manager.close()

    def _handle_shutdown(self, event, manager) -> None:
        pass

    def _handle_event(self, event, manager) -> None:
        self.last_events.append(event.headers)
    
    def _clean(self):
        self.last_events = []
    
    def _update_status(self):
        self.status = self._manager.status()
    
    def _send_change_to_broker(self, change_entities: list) -> None:

        for change_entity in change_entities:
            self._rabbitmq_producer.send(change_entity.get_dict())
    
    def changed_events_exists(self):
        return True if len(self.last_events) != 0 else False

    def update_events(self):
        change_entities = self._pabx.update(self.last_events)
        self._send_change_to_broker(change_entities)

    def run(self):
        self._logs_app.register('Service started')
        self._connect()

        try:
            # connect to the manager
            try:
                
                # register some callbacks
                self._manager.register_event(
                    'Shutdown', 
                    self._handle_shutdown
                )

                for evento in self.events:
                    self._manager.register_event(
                        evento, 
                        self._handle_event
                    )
                
                while True:
                    if self.changed_events_exists():
                        self.update_events()
                        self._clean()

                    self._update_status()
                    sleep(1)

            except ManagerSocketException as e:
                self._logs_app.register("Error connecting to the manager: %s" % e.strerror)
                sys.exit(1)

            except ManagerAuthException as e:
                self._logs_app.register("Error logging in to the manager: %s" % e.strerror)
                sys.exit(1)

            except ManagerException as e:
                self._logs_app.register("Error: %s" % e.strerror)
                sys.exit(1)

        finally:
            # remember to clean up
            self._logs_app.register('Encerrando conexões com o Manager')
            self._disconnect()


from ami2rabbitmq.repositories import Endpoints
from ami2rabbitmq.repositories import Queues
from ami2rabbitmq.repositories import QueueMembers
from ami2rabbitmq.repositories import Bridges
from ami2rabbitmq.repositories import QueueCallers


class Pabx():

    def __init__(self):
        self.endpoints = Endpoints()
        self.bridges = Bridges()
        self.queue_callers = QueueCallers()
        self.queues = Queues()
        self.queue_members = QueueMembers()
        self._changed_objects = []
        self._create_events = [
            'DeviceStateChange',
            'QueueMemberStatus'
        ]
        self._bridge_create_events = [
            'BridgeEnter',
            'QueueCallerJoin'
        ]
        self._bridge_delete_events = [
            'BridgeLeave',
            'QueueCallerLeave'
        ]
        self._endpoints_bridges_create_events = [
            'BridgeEnter',
        ]
        self._endpoints_bridges_delete_events = [
            'BridgeLeave',
        ]

    def _queue_state_change_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.queues.create(event)
        )
    
    def _endpoint_state_change_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.endpoints.create(event)
        )
    
    def _bridge_enter_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.bridges.create(event)
        )
    
    def _bridge_leave_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.bridges.delete(event)
        )
    
    def _queue_caller_leave_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.queue_callers.delete(event)
        )
    
    def _queue_caller_join_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.queue_callers.create(event)
        )

    def _device_state_change_event_update(self, event: dict):
        if 'Queue:' in event['Device']:
            self._queue_state_change_event_update(event)

        else:
            self._endpoint_state_change_event_update(event)
            
    def _queue_members_status_event_update(self, event: dict) -> None:
        self._changed_objects.append(
            self.queue_members.create(event)
        )
    
    def _create_events_update(self, event: dict) -> None:
        if event['Event'] == 'DeviceStateChange':
            self._device_state_change_event_update(event)
                
        elif event['Event'] == 'QueueMemberStatus':
            self._queue_members_status_event_update(event)
        
    def _bridge_create_events_update(self, event: dict) -> None:
        if event['Event'] == 'BridgeEnter':
            self._bridge_enter_event_update(event)
        
        elif event['Event'] == 'QueueCallerJoin':
            self._queue_caller_join_event_update(event)
    
    def _bridge_delete_events_update(self, event: dict) -> None:
        if event['Event'] == 'BridgeLeave':
            self._bridge_leave_event_update(event)
        
        elif event['Event'] == 'QueueCallerLeave':
            self._queue_caller_leave_event_update(event)
        
    def get_changed_objects(self) -> list:
        return self._changed_objects

    def clear(self) -> None:
        self.endpoints.clear()
        self.bridges.clear()
        self.queue_callers.clear()
        self.queues.clear()
        self.queue_members.clear()
        self._changed_objects = []

    def update(self, events: list) -> list:
        for event in events:
            if event['Event'] in self._create_events:
                self._create_events_update(event)
            
            elif event['Event'] in self._bridge_create_events:
                self._bridge_create_events_update(event)
            
            elif event['Event'] in self._bridge_delete_events:
                self._bridge_delete_events_update(event)
            
        changed_objects = self.get_changed_objects()
        self.clear()
        return changed_objects


def endpoint_parse(channel):
    posicao_final = channel.rfind('-')
    return channel[:posicao_final]

def queue_parse(queue):
    remover = ['Queue:']

    for item in remover:
        queue = queue.replace(item, '')
    
    partes_nome = queue.split('_')

    return f'{partes_nome[0]}_{partes_nome[1]}'


import json
import redis
from ami2rabbitmq.exceptions import InvalidKey


class RedisDB():
    _redis = redis.Redis()
    _key_prefix = 'Online:'

    def validate_key(self, key: str) -> bool:
        if not key.startswith(self._key_prefix):
            raise InvalidKey(f'Invalid Key: {key}')
        
        return True

    def get_all_keys(self) -> list:
        return self._redis.keys('Online:*')

    def exists(self, key: str) -> bool:
        return self._redis.exists(key) == 1
    
    def save(self, key: str, valor, ttl=None) -> bool:
        self.validate_key(key)
        return self._redis.set(key, json.dumps(valor), ttl)
    
    def get(self, key: str):
        if self.exists(key):
            return json.loads(self._redis.get(key))
        
        raise redis.exceptions.ResponseError(f'{key} not found')
    
    def delete(self, key: str) -> bool:

        if self.exists(key):
            return self._redis.delete(key)
        
        raise redis.exceptions.ResponseError(f'{key} not found')
    
    def flushall(self) -> None:
        for key in self.get_all_keys():
            self.delete(key)

redis_server = RedisDB()



from datetime import datetime
from ami2rabbitmq.utils import endpoint_parse, queue_parse
from ami2rabbitmq.exceptions import InvalidEvent


def get_endpoint_key(device):
    return f'Online:Endpoint:{device}'


def get_queue_key(device):
    return f'Online:Queue:{device}'


class Callerid():

    def __init__(self, num, name):
        self.num = num
        self.name = name

    def get_dict(self):
        return {
            'num' : self.num,
            'name' : self.name
        }


class Bridge():

    def __init__(self, event):
        self._bridge_events = [
            'BridgeEnter', 
            'BridgeLeave',
            'QueueCallerJoin',
            'QueueCallerLeave'
        ]

        if event['Event'] not in self._bridge_events:
            raise InvalidEvent(f"Event Invalid: {event['Event']}")

        self.last_update = datetime.now()
        self.id = event['BridgeUniqueid'] if 'BridgeUniqueid' in event else None
        self.device = endpoint_parse(event['Channel'])
        self.channel = event['Channel']
        self.state = event['ChannelState']
        self.callerid = Callerid(
            event['CallerIDNum'],
            event['CallerIDName']
        )
        self.connect_line = Callerid(
            event['ConnectedLineNum'],
            event['ConnectedLineName']
        )
        self.uniqueid = event['Uniqueid']
        self.linkedid = event['Linkedid']
        self.last_event = event['Event']
        self.type = 'bridge'
    
    def __str__(self):
        return self.id

    def get_key(self):
        return f'Online:Bridge:{self.uniqueid}'
    
    def get_dict(self):
        return {
            'id' : self.id,
            'device' : self.device,
            'channel' : self.channel,
            'state' : self.state,
            'last_event' : self.last_event,
            'last_update' : self.last_update.strftime('%Y-%m-%d %H:%M:%S'),
            'callerid' : self.callerid.get_dict(),
            'connect_line' : self.connect_line.get_dict(),
            'uniqueid' : self.uniqueid,
            'linkedid' : self.linkedid,
            'type' : self.type
        }


class QueueCaller(Bridge):
    
    def __init__(self, event):
        super().__init__(event)
        self.device = event['Queue']
        self.position = event['Position']
        self.count = event['Count']
        self.uniqueid = event['Uniqueid']
        self.linkedid = event['Linkedid']
        self.type = 'queue_caller'
    
    def __str__(self):
        return self.channel

    def get_key(self):
        return f'Online:QueueCaller:{self.device}:{self.uniqueid}'
    
    def get_dict(self):
        queue_caller_dict = super().get_dict()
        del queue_caller_dict['id']
        queue_caller_dict['position'] = self.position
        queue_caller_dict['count'] = self.count
        return queue_caller_dict
   

class Endpoint():

    def __init__(self, event):

        if event['Event'] != 'DeviceStateChange':
            raise InvalidEvent(f"Event Invalid: {event['Event']}")

        self.last_update = datetime.now()
        self.last_event = event['Event']
        self.device = event['Device']
        self.state = event['State']
        self.type = 'endpoint'
        self.bridges = []
    
    def __str__(self):
        return self.device
    
    def get_key(self):
        return get_endpoint_key(self.device)

    def get_dict(self):
        return {
            'device' : self.device,
            'state' : self.state,
            'last_update' : self.last_update.strftime('%Y-%m-%d %H:%M:%S'),
            'last_event' : self.last_event,
            'type' : self.type,
            'bridges' : self.bridges
        }


class QueueMember():

    def __init__(self, event):

        if event['Event'] != 'QueueMemberStatus':
            raise InvalidEvent(f"Event Invalid: {event['Event']}")

        self.last_update = datetime.now()
        self.last_event = event['Event']
        self.queue = event['Queue']
        self.member_name = event['MemberName']
        self.interface = event['Interface']
        self.state_interface = event['StateInterface']
        self.member_ship = event['Membership']
        self.penalty = event['Penalty']
        self.calls_taken = event['CallsTaken']
        self.last_call = event['LastCall']
        self.in_call = event['InCall']
        self.status = event['Status']
        self.paused = event['Paused']
        self.paused_reason = event['PausedReason']
        self.ring_inuse = event['Ringinuse']
        self.queues_list = []
        self.type = 'queue_member'
    
    def __str__(self) -> str:
        return f'{self.member_name}:{self.queue}'

    def get_key(self) -> str:
        return f'Online:QueueMember:{self.member_name}:{self.queue}'
    
    def get_list_queues_key(self) -> str:
        return f'Online:QueueMember:{self.member_name}'
    
    def get_dict(self):
        return {
            'queue' : self.queue,
            'member_name' : self.member_name,
            'interface' : self.interface,
            'state_interface' : self.state_interface,
            'member_ship' : self.member_ship,
            'penalty' : self.penalty,
            'calls_taken' : self.calls_taken,
            'last_call' : self.last_call,
            'in_call' : self.in_call,
            'status' : self.status,
            'paused' : self.paused,
            'paused_reason' : self.paused_reason,
            'ring_inuse' : self.ring_inuse,
            'last_update' : self.last_update.strftime('%Y-%m-%d %H:%M:%S'),
            'last_event' : self.last_event,
            'type' : self.type
        }


class Queue():

    def __init__(self, event):

        if event['Event'] != 'DeviceStateChange':
            raise InvalidEvent(f"Event Invalid: {event['Event']}")

        self.last_update = datetime.now()
        self.device = queue_parse(event['Device'])
        self.state = event['State']
        self.queue_callers = []
        self.last_event = event['Event']
        self.type = 'queue'
    
    def __str__(self):
        return self.device
    
    def get_key(self):
        return get_queue_key(self.device)

    def get_dict(self):
        return {
            'device' : self.device,
            'state' : self.state,
            'last_update' : self.last_update.strftime('%Y-%m-%d %H:%M:%S'),
            'queue_callers' : self.queue_callers,
            'last_event' : self.last_event,
            'type' : self.type
        }


import os
from setuptools import setup


lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = lib_folder + '/requirements.txt'
install_requires = []

# Obtain list of requirements
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()

# Obtain readme text for long description
with open("README.md", "r") as arq:
    readme = arq.read()

setup(
    name='ami2rabbitmq',
    version='0.1.5',
    license='MIT License',
    author='Tatianno Alves',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='tferreiraalves@gmail.com',
    keywords='asterisk ami rabbitmq producer',
    description=u'Producer Events Asterisk Manager Interface (AMI) for RabbitMQ Broker',
    packages=['ami2rabbitmq'],
    install_requires=install_requires,
)


from datetime import datetime


class LogsApp:

    def __init__(self, DEBUG):
        self.debug = DEBUG

    def format_msg(self, msg: str) -> str:
        return f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")} - {msg}'

    def register(self, msg: str) -> None:
        
        if self.debug:
            print(self.format_msg(msg))



import json
import pika


class RabbitMQProducer():
    __connection = None
    __channel = None
    

    def __init__(self, rabbitmq_settings, logs_app):
        self.host = rabbitmq_settings['host']
        self.username = rabbitmq_settings['user']
        self.password = rabbitmq_settings['password']
        self.queuename = rabbitmq_settings['queuename']
        self._logs_app = logs_app

    def connect(self):
        self.__connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=self.host,
                credentials=pika.PlainCredentials(
                    self.username, 
                    self.password    
                )
            )
        )
        self.__channel = self.__connection.channel()
        self.__channel.queue_declare(queue=self.queuename)

    def disconnect(self):
        self.__connection.close()
    
    def __rabbitmq_send__(self, mensagem):
        self.__channel.basic_publish(
            exchange='',
            routing_key=self.queuename,
            body=mensagem
        )
                       
    def send(self, mensagem):
        try:
            self.connect()
            self.__rabbitmq_send__(
                json.dumps(mensagem)
            )
            self.disconnect()
        
        except pika.exceptions.AMQPConnectionError as error:
            self._logs_app.register('Fail connection RabbitMQ')